import sys
import os
import json
import platform
import logging
import socket
import vcsa_utils
import uuid
import subprocess

from . import utils
from extensions import extend, Hook
from patch_specs import DiscoveryResult, ValidationResult, Question, Mismatch, \
    Requirements, PatchInfo, RequirementsResult
from l10n import msgMetadata as _T, localizedString as _
from reporting import getProgressReporter

log = logging.getLogger(__name__)
VMAFD_PAYLOAD_DIR = os.path.dirname(__file__)

# Import patches python directory where all individual python module reside
sys.path.append(os.path.join(VMAFD_PAYLOAD_DIR, 'patches'))
patches = [
]

def invoke_cmd(cmd):

    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    (output, error) = p.communicate()
    if error:
        print (error)
        return error
    else:
        return output

def getLDUInstallParam():
    cmd = b'install-parameter vmdir.ldu-guid'
    output = invoke_cmd(cmd)
    output = output.rstrip()
    return output

def setLDUInstallParam(guid):
    cmd = b'install-parameter vmdir.ldu-guid -s '
    cmd += guid
    output = invoke_cmd(cmd)
    log.info('set install parameter output: %s' % output)
    return output

def setLDURegistry(guid):
    vmdirKey = '[HKEY_THIS_MACHINE\\Services\\vmdir]'
    vmdirVal = 'LduGuid'
    vmafdKey = '[HKEY_THIS_MACHINE\\Services\\vmafd\\Parameters]'
    vmafdVal = 'LDU'

    utils = getUtil()

    ret = utils.setRegValue(vmdirKey, vmdirVal, guid)
    log.info('Set registry value in vmdir')
    if ret != 0:
        return ret

    ret = utils.setRegValue(vmafdKey, vmafdVal, guid)
    log.info('Set registry value in vmafd')

    return ret

def setLDU():
    guid = getLDUInstallParam()

    if (guid == None or not guid):
        log.info('There is no LDU-guid - making a new one')
        guid = str(uuid.uuid1())
        log.info('Setting LDU %s' % guid)
        setLDUInstallParam(guid)
    else:
        log.info('LDU-guid already exists: %s' % guid)
    log.info('Setting LDU to VMAFD')

    result = setLDURegistry(guid)
    if (result != 0):
        raise Exception('Setting the LDU Failed')
    return

def getUtil():
    if os.name == 'nt':
        return utils.PatchUtilsWin()
    else:
        return utils.PatchUtilsLin()

def getCurrentVersion():
    return getUtil().getSourceVersion()

def getLatestVersion():
    if len(patches) > 0:
        return patches[-1][0]
    return '6.5.0.0'

def getChangesSummary():
    latestPatchScript = patches[-1][1]
    mod = __import__(latestPatchScript)
    changesSummary = None
    if hasattr(mod, 'getChanges'):
        changesSummary = mod.getChanges()

    return changesSummary

@extend(Hook.Discovery)
def discover(ctx):
    '''DiscoveryResult discover(PatchContext ctx) throw UserError'''
    if platform.system().lower() == 'windows':
        # I don't want to take part in windows patching workflow
        return None

    return vcsa_utils.getComponentDiscoveryResult('vmafdd')

@extend(Hook.Requirements)
def collectRequirements(ctx):
    '''RequirementsResult collectRequirements(PatchContext ctx)'''
    mismatches = []
    requirements = Requirements()
    patchInfo = PatchInfo()

    try:
        requirements = Requirements(requiredDiskSpace={'/storage/core' :
                                                           getUtil().getRequiredDiskSpace()},
                                    questions=[],
                                    rebootRequired=False)
    except Exception: #pylint: disable=W0703
        log.exception('Cannot connect to VMWare Afd Services')
        mismatches.append(Mismatch(text=_(_T('vmafd.error.text',
                                             'Unable to connect to VMWare Afd Services')),
                                   description=_(_T('vmafd.error.description',
                                                    'VMware Afd Service is down')),
                                   resolution=_(_T('vmafd.error.resolution',
                                                   'Please check that VMware Afd Service is running')),
                                   severity=Mismatch.ERROR))

    return RequirementsResult(requirements, patchInfo, mismatches)

@extend(Hook.Validation)
def validate(ctx):
    '''ValidationResult validate(PatchContext ctx)'''
    mismatches = []
    return ValidationResult(mismatches)

@extend(Hook.Prepatch)
def prePatch(ctx): #pylint: disable=W0613
    '''void prePatch(PatchContext sharedCtx) throw UserUpgradeError'''

def _doIncrementalPatching(ctx):
    utils = getUtil()
    currVersion = getCurrentVersion()
    log.info('Vmafd current version: %s', currVersion)
    patchCurrentVersion = currVersion is None

    for ver, modulePath in patches:
        compResult = utils.versionCompare(currVersion, ver)
        patchCurrentVersion |= compResult < 0

        if patchCurrentVersion:
            log.info('Applying patch %s on version %s' % (ver, modulePath))
            mod = __import__(modulePath)
            mod.doPatching(ctx)
            log.info('Patch %s applied' % (modulePath))

    utils.setSourceVersion(getLatestVersion())

@extend(Hook.Patch)
def patch(ctx):
    '''void patch(PatchContext sharedCtx) throw UserUpgradeError'''
    setLDU()

    utils = getUtil()

    progressReporter = getProgressReporter()

    progressReporter.updateProgress(0, _(_T('vmafd.patch.begin',
                                            'Start VMWare Afd Services patching')))
    LwIsStarted = utils.start_lwsmd()

    _doIncrementalPatching(ctx)

    if LwIsStarted:
        utils.stop_lwsmd()

    progressReporter.updateProgress(100, _(_T('vmafd.patch.complete',
                                              'Patching VMWare Afd Services completed')))
